/*
 *"R & B": Real-coded GAs & Bit-string GAs
 *
 * immigration module
 */

#include "randb.h"
#include "utl_rand.h"
#include "utl_alloc.h"
#include "logUtil.h"

static int *randomRing = NULL;

/*
 * Prottype declarations of a static functions
 */

/*-----
 * select_emigrants
 */
static int select_emigrants(const GaParam_T *ga_param, GaIsland_T *island);

/*-----
 * shuffle_randomring
 */
static int shuffle_randomring(int size);

/*-----
 * random_swap_immigration
 */
static int random_swap_immigration(const GaParam_T *ga_param, GaIsland_T *islands);

/*-----
 * random_ring_immigration
 */
static int random_ring_immigration(const GaParam_T *ga_param, GaIsland_T *islands);

/***********************
 * Public functions    *
 ***********************/

/*-----
 * MakeImmigrationBuffer
 */
int MakeImmigrationBuffer(const GaParam_T *ga_param, GaIsland_T *islands)
{
	int		i;
	int		rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("MakeImmigrationBuffer enter.\n"));
#endif /* DBG */

	if(ga_param->gaTypeParam.immigrationSwitch == RANDOM_RING){
		for(i = 0; i < ga_param->gaTypeParam.numIsland; i++){
			if((islands[i].emigrants = 
				IntVectorAlloc(ga_param->gaTypeParam.numImmigrants)) == NULL){
				goto out;
			}
			if((islands[i].immigrants =
				GaGeneVectorAlloc(ga_param->gaTypeParam.numImmigrants)) == NULL){
				goto out;
			}
		}

		if((randomRing = 
			IntVectorAlloc(ga_param->gaTypeParam.numIsland * 2)) == NULL){
			goto out;
		}
		for(i = 0; i < 2 * ga_param->gaTypeParam.numIsland; i++){
			randomRing[i] = (i + 1) % ga_param->gaTypeParam.numIsland;
		}	
	}
	
	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("MakeImmigrationBuffer exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * FreeImmigrationBuffer
 */
void FreeImmigrationBuffer(const GaParam_T *ga_param, GaIsland_T *islands)
{
	int	i;

#ifdef DBG
	DBGW(10, ("FreeImmigrationBuffer enter.\n"));
#endif /* DBG */

	if(ga_param->gaTypeParam.immigrationSwitch == RANDOM_RING){
		if(islands != NULL){
			for(i = 0; i < ga_param->gaTypeParam.numIsland; i++){
				if(islands[i].emigrants != NULL){
					FreeIntVector(islands[i].emigrants);
				}
				if(islands[i].immigrants != NULL){
					FreeGaGeneVector(islands[i].immigrants);
				}
			}
		}
	}

	if(randomRing != NULL){
		FreeIntVector(randomRing);
	}

#ifdef DBG
	DBGW(10, ("FreeImmigrationBuffer exit\n"));
#endif /* DBG */

}	

/*-----
 * Immigrateؿ
 *      ֤ΰܽԤ
 *      ܽˡ
 *      ܽˡ
 *      ͡EXIT_NORMAL
 *      GaParam_T*(I), GaGene_T**(I/O)
 */
int Immigrate(const GaParam_T *ga_param, GaIsland_T *islands)
{
	int	rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("Immigrate enter.\n"));
#endif /* DBG */
		
	switch(ga_param->gaTypeParam.immigrationSwitch){
	case RANDOM_SWAP:
		if(random_swap_immigration(ga_param, islands) == EXIT_ERROR){
			goto out;
		}
		break;
	case RANDOM_RING:
		if(random_ring_immigration(ga_param, islands) == EXIT_ERROR){
			goto out;
		}
		break;
	default:
		log_sys(" ## System Error: In Immigrate.\n");
		goto out;
	}

	rv = EXIT_NORMAL;
out:

#ifdef DBG
	DBGW(10, ("Immigrate exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/***********************
 * Static functions    *
 ***********************/

/*-----
 * select_emigrants
 */
static int select_emigrants(const GaParam_T *ga_param, GaIsland_T *island)
{
	int 	i, j;
	int		num_elite;
	int		island_pop;
	int		candidate;
	int		rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("select_emigrants enter.\n"));
#endif /* DBG */
	
	num_elite = ga_param->digenesisParam.numElite;
	island_pop = ga_param->gaTypeParam.islandPopulation;

	if(ga_param->gaTypeParam.numImmigrants > island_pop - num_elite){
		log_sys(" ## Error: Immigration rate too large.\n");
		goto out;
	}
	
	for(i = 0; i < ga_param->gaTypeParam.numImmigrants;){
		candidate = GenIntRand(num_elite, island_pop);
		for(j = 0; j < i; j++){
			if(island->emigrants[j] == island->sortedIndex[candidate]){
				continue;
			}
		}
		island->emigrants[i] = island->sortedIndex[candidate];
		i++;
	}

	rv = EXIT_NORMAL;
out:

#ifdef DBG
	DBGW(10, ("select_emigrants exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
	

/*-----
 * shuffle_randomring
 */
static int shuffle_randomring(int size)
{
	int		i;
	int		*work;

#ifdef DBG
	DBGW(10, ("shuffle_randomring enter.\n"));
#endif /* DBG */

	work = randomRing + size;

	for(i = 0; i < size;){
		int a;
		int b;
		int temp;

		a = GenIntRand(0, size);
		b = GenIntRand(0, size);

		if(work[a] == b || work[b] == a) continue;

		temp = work[a];
		work[a] = work[b];
		work[b] = temp;

		i++;
	}

	for(i = 0; i < size; i++){
		randomRing[work[i]] = work[(i + 1) % size];
	}

#ifdef DBG
	DBGW(10, ("shuffle_randomring exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
	

		
static int random_swap_immigration(const GaParam_T *ga_param, GaIsland_T *islands){
	int 	i;
	int 	island_id1, island_id2;
	int		gene_id1, gene_id2;
	int		num_island;
	int		island_pop;
	int		num_elite;
	GaGene_T 	temp;

#ifdef DBG
	DBGW(10, ("random_swap_immigration enter.\n"));
#endif /* DBG */
	
	num_island = ga_param->gaTypeParam.numIsland;
	island_pop = ga_param->gaTypeParam.islandPopulation;
	num_elite = ga_param->digenesisParam.numElite;
	
	
	for(i = 0; i < ga_param->gaTypeParam.totalPopulation; i++){
		if(GenUniformRand(0, 1) < ga_param->gaTypeParam.immigrationRate){
			island_id1 = GenIntRand(0, num_island);
			island_id2 = GenIntRand(0, num_island);
			gene_id1 = GenIntRand(num_elite, island_pop);
			gene_id2 = GenIntRand(num_elite, island_pop);
			gene_id1 = islands[island_id1].sortedIndex[gene_id1];
			gene_id2 = islands[island_id2].sortedIndex[gene_id2];
			temp = islands[island_id1].residents[gene_id1];
			islands[island_id1].residents[gene_id1] =
				islands[island_id2].residents[gene_id2];
			islands[island_id2].residents[gene_id2] = temp;
		}
	}

	for(i = 0; i < num_island; i++){
		SortByFitness(&(islands[i]));
	}

#ifdef DBG
	DBGW(10, ("random_swap_immigration exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}

/*-----
 * random_ring_immigration
 */
static int random_ring_immigration(const GaParam_T *ga_param, GaIsland_T *islands)
{
	int i, j;
	int num_island;
	int	num_immigrants;
	int	rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("ramdo_ring_immigration enter.\n"));
#endif /* DBG */
	
	num_island = ga_param->gaTypeParam.numIsland;
	num_immigrants = ga_param->gaTypeParam.numImmigrants;
	
	shuffle_randomring(num_island);

	for(i = 0; i < num_island; i++){
		if(select_emigrants(ga_param, &(islands[i])) == EXIT_ERROR){
			goto out;
		}
		for(j = 0; j < num_immigrants; j++){
			islands[randomRing[i]].immigrants[j]
				= islands[i].residents[islands[i].emigrants[j]];
		}
	}

	for(i = 0; i < num_island; i++){
		for(j = 0; j < num_immigrants; j++){
			islands[i].residents[islands[i].emigrants[j]]
				= islands[i].immigrants[j];
		}
	}

	for(i = 0; i < num_island; i++){
		SortByFitness(&(islands[i]));
	}
	
	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("random_ring_immigration exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}
