#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "randb.h"
#include "utl_alge.h"
#include "utl_alloc.h"
#include "logUtil.h"

#ifdef _TEST_
void dump_matrix(char *s,double **in_mat, int num_row, int num_col);
#endif

/*-----
 * VectorInnerProduct
 */
double VectorInnerProduct(double *in_vec1, double *in_vec2, int dim_vec)
{
	int	i;
	double	rv = 0;

#ifdef DBG
	DBGW(10, ("VectorInnerProduct enter.\n"));
#endif /* DBG */

	for(i = 0; i < dim_vec; i++){
		rv += in_vec1[i] * in_vec2[i];
	}

#ifdef DBG
	DBGW(10, ("VectorInnerProduct exit(%f)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * VectorNorm
 */
double VectorNorm(double *in_vec, int dim_vec)
{
	double 	rv;

#ifdef DBG
	DBGW(10, ("VectorNorm enter.\n"));
#endif /* DBG */

	rv = VectorInnerProduct(in_vec, in_vec, dim_vec);

#ifdef DBG
	DBGW(10, ("VectorNorma exit(%f)\n", rv));
#endif /* DBG */

	return sqrt(rv);
}

/*-----
 * CalculateCenterOfGravity
 */
int CalculateCenterOfGravity(double **in_vecs, int num_vec, int dim_vec, double *out_vec)
{
        int     i, j;

#ifdef DBG
	DBGW(10, ("CalcurateCenterOfGravity enter.\n"));
#endif /* DBG */

        for(i = 0; i < dim_vec; i++){
                out_vec[i] = 0;
                for(j = 0; j < num_vec; j++){
                        out_vec[i] += in_vecs[j][i];
                }
                out_vec[i] = out_vec[i]/(num_vec);
        }

#ifdef DBG
	DBGW(10, ("CalcurateCenterOfGravity exit(0)\n"));
#endif /* DBG */

        return EXIT_NORMAL;
}

/*-----
 * CopyMatrix
 */
int CopyMatrix(double **in_mat, int num_row, int num_col, double **out_mat)
{
	int		i,j;

#ifdef DBG
	DBGW(10, ("CopyMatrix enter.\n"));
#endif /* DBG */

	for(i = 0; i < num_row; i++){
		for(j = 0; j < num_col; j++){
			out_mat[i][j] = in_mat[i][j];
		}
	}

#ifdef DBG
	DBGW(10, ("CopyMatrix exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
		

/*-----
 * TransposedMatrix
 */
int TransposedMatrix(double **in_mat, int num_row, int num_col, double **out_mat)
{
	int	i,j;

#ifdef DBG
	DBGW(10, ("TranposedMatrix enter.\n"));
#endif /* DBG */

	for(i = 0; i < num_row; i++){
		for(j = 0; j < num_col; j++){
			out_mat[j][i] = in_mat[i][j];
		}
	}

#ifdef DBG
	DBGW(10, ("TransposedMatrix exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
	
/*-----
 * ProductMatrix
 */
int ProductMatrix(double **in_mat1, int num_row1, int num_col1, double **in_mat2, int num_row2, int num_col2, double **out_mat)
{
	int 	i, j, k;
	int		rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("ProductMatrix enter.\n"));
#endif /* DBG */

	if(num_col1 != num_row2){
		log_sys(" ## Error: Cannot calculate product of matrices.\n");
		goto out;
	}

	for(i = 0; i < num_row1; i++){
		for(j = 0; j < num_col2; j++){
			out_mat[i][j] = 0;
			for(k = 0; k < num_col1; k++){
				out_mat[i][j] += in_mat1[i][k] * in_mat2[k][j];
			}
		}
	}

	rv = EXIT_NORMAL;

out:
#ifdef DBG
	DBGW(10, ("ProductMatrix exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

/*-----
 * ReductMatrix
 */
int ReductMatrix(double **in_mat, int num_row, int num_col, double **out_mat)
{
	int		i, j, k;
	int		pivot_row, pivot_col;
	int		rank = 0;
	double	max;
	double	temp;

#ifdef DBG
	DBGW(10, ("ReductMatrix enter.\n"));
#endif /* DBG */

	CopyMatrix(in_mat, num_row, num_col, out_mat);
	
	for(i = 0; i < num_row; i++){
		max = 0.0;
		for(j = i; j < num_row; j++){
			for (k = i; k < num_col; k++){
			 	if(max < fabs(out_mat[j][k])){
					max = fabs(out_mat[j][k]);
					pivot_row = j;
					pivot_col = k;
				}
			}
		}
#ifdef _TEST_
		printf("%d try: pivot = X[%d][%d]\n", rank, pivot_row, pivot_col);

		dump_matrix("X(in)", out_mat, 3, 4);
#endif		
		if (max <= 0.0) goto out;

		if(i != pivot_row){
			for(j = 0; j < num_col; j++){
				temp = out_mat[pivot_row][j];
				out_mat[pivot_row][j] = out_mat[i][j];
				out_mat[i][j] = temp;
			}
		}

		if(i != pivot_col){
			for(j = 0; j < num_row; j++){
				temp = out_mat[j][pivot_col];
				out_mat[j][pivot_col] = out_mat[j][i];
				out_mat[j][i] = temp;
			}
		}

#ifdef _TEST_
		dump_matrix("X(out)", out_mat, 3, 4);
#endif
		temp = out_mat[i][i];
		for(j = i; j < num_col; j++){
			out_mat[i][j] /= temp;
		}
#ifdef _TEST_
		dump_matrix("X(out2)", out_mat, 3, 4);
#endif
		
		for(j = i + 1; j < num_row; j++){
			temp = out_mat[j][i];
			for(k = i; k < num_col; k++){
				out_mat[j][k] -= temp * out_mat[i][k];
			}
		}
		rank++;
	}

out:
#ifdef DBG
	DBGW(10, ("ReductMatrix exit(%d)\n", rank));
#endif /* DBG */

	return rank;
}


/*-----
 * GrammSchmidtOrthogonalize
 */
int GrammSchmidtOrthogonalize(double **in_vecs, int num_vec, int dim_vec, double **out_vecs)
{
	int		i, j, k;
	double	norm;

#ifdef DBG
	DBGW(10, ("GrammSchmidtOrthogonalize enter.\n"));
#endif /* DBG */

	/* in_vecs[0] -> out_vecs[0] */
	memcpy(out_vecs[0], in_vecs[0], dim_vec * sizeof(double));

	norm = VectorNorm(out_vecs[0], dim_vec);
	for(i = 0; i < dim_vec; i++){
		out_vecs[0][i] /= norm;
	}
	
	for(i = 1; i < num_vec; i++){
		for(j = 0; j < dim_vec; j++){
			out_vecs[i][j] = in_vecs[i][j];
			for(k = 0; k < i; k++){
				out_vecs[i][j] -=
					VectorInnerProduct(in_vecs[i], out_vecs[k], dim_vec) * out_vecs[k][j];
			}
		}
		norm = VectorNorm(out_vecs[i], dim_vec);
		for(j = 0; j < dim_vec; j++){
			out_vecs[i][j] /= norm;
		}
	}

#ifdef DBG
	DBGW(10, ("GrammSchmidtOrthogonalize exit(0)\n"));
#endif /* DBG */

	return EXIT_NORMAL;
}
