#include <stdio.h>
#include <math.h>

#include "utl_encode.h"
#include "logUtil.h"
#include "randb.h"

/*
 * static functions
 */

static int b_encode(int in_val, char *bit_array, int size)
{
	int i;
	int val;
	int rv = EXIT_ERROR;
	
#ifdef DBG
	DBGW(10, ("b_encode enter.\n"));
#endif /* DBG */
	
	if(in_val >= (int)pow(2, size)){
		log_sys(" ## Error: Too large value to binary encode.\n");
		goto out;
	}

	val = in_val;
	for(i = 0; i < size; i++){
		if(val & (1 << i)) {
			bit_array[size-1-i] = 1;
		} else {
			bit_array[size-1-i] = 0;
		}
	}
		
	rv = EXIT_NORMAL;
out:	
#ifdef DBG
	DBGW(10, ("b_encode exit(%d).\n"));
#endif /* DBG */

	return rv;
}
	
static int b_decode(const char *bit_array, int size)
{
	int	i;
	int rv = 0;

#ifdef DBG
	DBGW(10, ("b_decode enter.\n"));
#endif /* DBG */

	for(i = 0; i < size; i++){
		if(bit_array[i]){
			rv += 1 << (size - i - 1);
		}
	}

#ifdef DBG
	DBGW(10, ("b_decode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

static int g_encode(int in_val, char *bit_array, int size)
{
	int i;
	int f_swap = 0;
	int mask;
	
	int rv = EXIT_ERROR;
	
#ifdef DBG
	DBGW(10, ("g_encode enter.\n"));
#endif /* DBG */
	
	if(in_val >= (int)pow(2, size)){
		log_sys(" ## Error: Too large value to gray encode.\n");
		goto out;
	}

	for(mask = 1 << (sizeof(int) * 8 - 2); mask != 0; mask >>= 1){
		if(f_swap){
			in_val ^= mask;
			if(in_val & mask) f_swap = 0;
		} else {
			if(in_val & mask) f_swap = 1;
		}
	}

	for(i = 0; i < size; i++, in_val >>= 1){
		bit_array[size-i-1] = in_val & 1;
	}
	
	rv = EXIT_NORMAL;
out:	
#ifdef DBG
	DBGW(10, ("g_encode exit(%d).\n"));
#endif /* DBG */

	return rv;
}

static int g_decode(const char *bit_array, int size)
{
	int i, mask;
	int	rv = 0;
	int f_swap = 0;

#ifdef DBG
	DBGW(10, ("g_decode enter.\n"));
#endif /* DBG */

	for(i = 0; i < size; i++){
		if(bit_array[i]){
			rv += 1 << (size - i - 1);
		}
	}
	
	for(mask = 1 << (size - 1); mask != 0; mask >>= 1){
		if(f_swap){
			if(rv & mask) f_swap = 0;
			rv ^= mask;
		} else if(rv & mask){
			f_swap = 1;
		}
	}

#ifdef DBG
	DBGW(10, ("g_decode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

int BinaryEncode(const int *in_array, int in_array_size, char *bit_array, int b_array_size, int quantumNumber)
{
	int i;
	int rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("BinaryEncode enter.\n"));
#endif /* DBG */
	
	if(in_array_size * quantumNumber != b_array_size){
		log_sys(" ## Error: Array size mismatched(in BinaryEncode).\n");
		goto out;
	}
		
	for(i = 0; i < in_array_size; i++){
		if(b_encode(in_array[i], bit_array + i * quantumNumber, quantumNumber)
					== EXIT_ERROR){
			goto out;
		}	
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("BinaryEncode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

int BinaryDecode(const char *bit_array, int b_array_size, int quantumNumber, int out_array_size, int *out_array)
{
	int i,j;
	int	rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("BinaryDecode enter.\n"));
#endif /* DBG */

	if(quantumNumber * out_array_size != b_array_size){
		log_sys(" ## Error: Array size mismatched(in BinaryDecode).\n");
		goto out;
	}

	for(i = j = 0; i < b_array_size; i += quantumNumber, j++){
		out_array[j] = b_decode(bit_array + i, quantumNumber);
	}

	rv = EXIT_NORMAL;
out:

#ifdef DBG
	DBGW(10, ("BinaryDecode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

int GrayEncode(const int *in_array, int in_array_size, char *bit_array, int b_array_size, int quantumNumber)
{
	int i;
	int rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("GrayEncode enter.\n"));
#endif /* DBG */

	if(in_array_size * quantumNumber != b_array_size){
		log_sys(" ## Error: Array size mismatched(in GrayEncode).\n");
		goto out;
	}
		
	for(i = 0; i < in_array_size; i++){
		if(g_encode(in_array[i], bit_array + i * quantumNumber, quantumNumber)
					== EXIT_ERROR){
			goto out;
		}	
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("GrayEncode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}


int GrayDecode(const char *bit_array, int b_array_size, int quantumNumber, int out_array_size, int *out_array)
{
	int i, j;
	int	rv = EXIT_ERROR;

#ifdef DBG
	DBGW(10, ("GrayDecode enter.\n"));
#endif /* DBG */
	
	if(quantumNumber * out_array_size != b_array_size){
		log_sys(" ## Error: Array size mismatched(in GrayDecode).\n");
		goto out;
	}

	for(i = j = 0; i < b_array_size; i += quantumNumber, j++){
		out_array[j] = g_decode(bit_array + i, quantumNumber);
	}

	rv = EXIT_NORMAL;
out:
#ifdef DBG
	DBGW(10, ("GrayDecode exit(%d)\n", rv));
#endif /* DBG */

	return rv;
}

#ifdef _TEST_

int main()
{
#define BL 100
#define IL 10
	char 	bit[BL];
	int		seisu[IL];
	int		seisu2[IL];
	int		qn = 10;
	int		i;
	
	for(i = 0; i < IL; i++){
		seisu[i] = 3 + i * 5;
	}

	BinaryEncode(seisu, IL, bit, BL, qn);

	for(i = 0; i < BL;i++){
		printf("%d", bit[i]);
		if( (i + 1) % qn == 0) printf("\n");
	}
	
	BinaryDecode(bit, BL, qn, IL, seisu2);

	for(i = 0; i < IL; i++){
		printf("%d,", seisu2[i]);
	}
	printf("\n");
	
	return 0;
}

#endif /* _TEST_ */
